import os
import abc
import typing as t
from omegaconf import (OmegaConf, DictConfig)


class MLCubeInstance(abc.ABC):
    """ Base class for different instantiations of MLCube (local, remote, directory, archive, container ...). """
    @abc.abstractmethod
    def uri(self) -> t.Text:
        raise NotImplementedError


MLCubeInstanceType = t.TypeVar('MLCubeInstanceType', bound=MLCubeInstance)


class MLCubeDirectory(MLCubeInstance):
    """ An MLCube instantiation as a local directory. """
    def __init__(self, path: t.Optional[t.Text] = None) -> None:
        if path is None:
            path = os.getcwd()
        path = os.path.abspath(path)
        if os.path.isfile(path):
            self.path, self.file = os.path.split(path)
        else:
            self.path, self.file = os.path.abspath(path), 'mlcube.yaml'

    def uri(self) -> t.Text:
        return os.path.join(self.path, self.file)


class CliParser(object):
    @staticmethod
    def parse_mlcube_arg(mlcube: t.Optional[t.Text]) -> MLCubeInstanceType:
        """ Parse value of the `--mlcube` command line argument.
        Args:
            mlcube: Path to a MLCube directory or `mlcube.yaml` file. If it's a directory, standard name
                `mlcube.yaml` is assumed for MLCube definition file.
        Returns:
            One of child classes of MLCubeInstance that represents this MLCube.
        """
        return MLCubeDirectory(mlcube)

    @staticmethod
    def parse_list_arg(arg: t.Optional[t.Text], default: t.Optional[t.Text] = None) -> t.List[t.Text]:
        """ Parse a string into list of strings using `,` as a separator.
        Args:
            arg: String if elements separated with `,`.
            default: Default value for `arg` if `arg` is None or empty.
        Returns:
            List of items.
        """
        arg = arg or default
        if not arg:
            return []
        return arg.split(',')

    @staticmethod
    def parse_extra_arg(*args: t.Text) -> t.Tuple[DictConfig, t.Dict]:
        """ Parse extra arguments on a command line.
        These arguments correspond to:
            - MLCube runtime arguments. These start with `-P` prefix and are translated to a nested dictionaries
                structure using `.` as a separator. For instance, `-Pdocker.image=mlcommons/mnist:0.0.1` translates to
                python dictionary {'docker': {'image': 'mlcommons/mnist:0.0.1'}}.
            - Task arguments are all other arguments that do not star with `-P`. These arguments are input/output
                arguments of tasks.
        Args:
            args: List of arguments that have not been parsed before.
        Returns:
            Tuple of two dictionaries: (mlcube_arguments, task_arguments).
        """
        mlcube_args = OmegaConf.from_dotlist([arg[2:] for arg in args if arg.startswith('-P')])

        task_args = [arg.split('=') for arg in args if not arg.startswith('-P')]
        task_args = {arg[0]: arg[1] for arg in task_args}

        return mlcube_args, task_args



import os
import logging
import typing as t
from pathlib import Path
from omegaconf import (OmegaConf, DictConfig)
from mlcube.errors import MLCubeError
from mlcube.platform import Platform
from mlcube.runner import Runner

logger = logging.getLogger(__name__)


class SystemSettings(object):

    @staticmethod
    def system_settings_file() -> t.Text:
        return os.path.abspath(os.environ.get('MLCUBE_SYSTEM_SETTINGS', Path.home() / 'mlcube.yaml'))

    def __init__(self, path: t.Optional[t.Text] = None) -> None:
        self.path: Path = Path(path if path is not None else SystemSettings.system_settings_file()).resolve()
        if not self.path.exists():
            logger.info("MLCube system settings file does not exist (%s).", str(self.path))
            self.path.touch()
        else:
            logger.info("MLCube system settings file exists (%s)", str(self.path))
        self.settings: DictConfig = OmegaConf.load(self.path)
        if not isinstance(self.settings, DictConfig):
            raise ValueError(f"Invalid object read from {self.path} (type = {type(self.settings)}). "
                             f"Expecting DictConfig.")

        updated = False
        for key in ('runners', 'platforms', 'storage'):
            if key not in self.settings:
                updated = True
                self.settings[key] = {}
        if updated:
            self.save()

    @property
    def runners(self) -> DictConfig:
        return self.settings.runners

    @property
    def platforms(self) -> DictConfig:
        return self.settings.platforms

    @property
    def storage(self) -> DictConfig:
        return self.settings.storage

    def save(self, resolve: bool = False) -> 'SystemSettings':
        OmegaConf.save(self.settings, self.path, resolve=resolve)
        return self

    def update_installed_runners(self) -> 'SystemSettings':
        installed_runners: t.Dict = Platform.get_installed_runners()
        updated: bool = False
        for platform_name, platform_spec in installed_runners.items():
            if platform_name not in self.settings.runners:
                updated = True
                self.settings.runners[platform_name] = platform_spec['config']
            if platform_name not in self.settings.platforms:
                updated = True
                self.settings.platforms[platform_name] = platform_spec['runner_cls'].CONFIG.DEFAULT
        if updated:
            self.save()
        return self

    def get_platform(self, platform: t.Optional[t.Text] = None) -> DictConfig:
        if not platform:
            return OmegaConf.create({})
        return self.settings.platforms.get(platform, OmegaConf.create({}))

    def create_platform(self, args: t.Optional[t.Tuple[t.Text, t.Text]] = None) -> 'SystemSettings':
        if args is None:
            return self
        runner, platform = args
        if platform in self.settings.platforms:
            raise MLCubeError(f"Platform ({platform}) already exists ({self.settings.platforms[platform]}).")
        if runner not in self.settings.runners:
            raise MLCubeError(f"Unknown runner ({runner}). Installed runners = {str(self.settings.runners.keys())}")
        runner_cls: t.Type[Runner] = Platform.get_runner(self.runners.get(runner, None))
        self.settings.platforms[platform] = runner_cls.CONFIG.DEFAULT
        self.save()
        return self

    def remove_platform(self, platform: t.Optional[t.Text] = None) -> 'SystemSettings':
        if platform and platform in self.settings.platforms:
            del self.settings.platforms[platform]
            self.save()
        return self

    def copy_platform(self, args: t.Optional[t.Tuple[t.Text, t.Text]] = None,
                      delete_source: bool = False) -> 'SystemSettings':
        if args is None:
            return self
        old_name, new_name = args
        if old_name not in self.settings.platforms:
            raise MLCubeError(f"Platform ({old_name}) does not exist.")
        if new_name in self.settings.platforms:
            raise MLCubeError(f"Platform ({new_name}) already exists ({self.settings.platforms[new_name]}).")
        self.settings.platforms[new_name] = self.settings.platforms[old_name]
        if delete_source:
            del self.settings.platforms[old_name]
        self.save()
        return self

    def rename_runner(self, args: t.Optional[t.Tuple[t.Text, t.Text]] = None,
                      update_platforms: bool = False) -> 'SystemSettings':
        if args is None:
            return self
        old_name, new_name = args
        if old_name not in self.settings.runners:
            raise MLCubeError(f"Runner ({old_name}) does not exist.")
        if new_name in self.settings.runners:
            raise MLCubeError(f"Runner ({new_name}) already exists ({self.settings.runners[new_name]}).")

        for platform_name, platform_def in self.settings.platforms:
            if platform_def.runner == old_name:
                if not update_platforms:
                    raise MLCubeError(
                        f"Cannot rename runner ({old_name}). Some platforms reference this runner, and need to "
                        "be updated, but update_platforms is false (if you are running an mlcube from CLI, "
                        "provide `--update-platforms` switch)."
                    )
                platform_def.runner = new_name

        self.settings.runners[new_name] = self.settings.runners[old_name]
        del self.settings.runners[old_name]
        self.save()
        return self

    def remove_runner(self, runner: t.Optional[t.Text] = None, remove_platforms: bool = False) -> 'SystemSettings':
        if not runner or runner not in self.settings.runners:
            return self

        platforms = [p for p in self.settings.platforms if self.settings.platforms[p].runner == runner]
        if platforms and not remove_platforms:
            raise MLCubeError(
                f"Cannot remove runner ({runner}). Some platforms reference this runner, and need to "
                "be removed, but remove_platforms is false (if you are running an mlcube from CLI, "
                "provide `--remove-platforms` switch)."
            )
        for platform in platforms:
            del self.settings.platforms[platform]
        del self.settings.runners[runner]
        self.save()
        return self



import pkgutil
import logging
import importlib
import typing as t
from omegaconf import DictConfig
from mlcube.runner import Runner

logger = logging.getLogger(__name__)

__all__ = ['Platform']


class Platform(object):
    reference_runners = {
        'docker': dict(pkg='mlcube_docker'),
        'singularity': dict(pkg='mlcube_singularity'),
        'ssh': dict(pkg='mlcube_ssh'),
        'gcp': dict(pkg='mlcube_gcp'),
        'k8s': dict(pkg='mlcube_k8s'),
        'kubeflow': dict(pkg='mlcube_kubeflow'),
    }

    @staticmethod
    def get_installed_runners() -> t.Dict:
        candidate_runners = {
            name: importlib.import_module(name) for _, name, _ in pkgutil.iter_modules() if name.startswith('mlcube_')
        }
        installed_runners = {}
        for pkg_name, module in candidate_runners.items():
            try:
                get_runner_class: t.Callable = getattr(module, 'get_runner_class')
                runner_cls: t.Type[Runner] = get_runner_class()
                if not issubclass(runner_cls, Runner):
                    raise TypeError(f"Invalid type of a runner ({runner_cls}). Expecting subclass of {Runner}.")
                runner_name = runner_cls.CONFIG.DEFAULT.runner
                installed_runners[runner_name] = dict(config=dict(pkg=pkg_name), runner_cls=runner_cls)
                logger.info("Found installed MLCube runner: platform=%s, pkg=%s", runner_name, pkg_name)
            except (AttributeError, TypeError) as e:
                logger.warning("Package (%s) is not a valid MLCube runner. Error = \"%s\"", pkg_name, str(e))
        return installed_runners

    @staticmethod
    def get_runner(runner_config: t.Optional[DictConfig]) -> t.Type[Runner]:
        if not runner_config:
            raise RuntimeError("Can't create runner. Runner config is null or empty.")
        if 'pkg' not in runner_config:
            raise RuntimeError(f"Do not know how to instantiate a runner. Runner config={str(runner_config)}")
        module = importlib.import_module(runner_config.pkg)
        get_runner_class: t.Callable = getattr(module, 'get_runner_class')
        return get_runner_class()



"""
This requires the MLCube 2.0 that's located somewhere in one of dev branches.
"""
import os
import sys
import click
import logging
import coloredlogs
import typing as t
from omegaconf import (OmegaConf, DictConfig)
from mlcube.config import MLCubeConfig
from mlcube.errors import (IllegalParameterValueError, MLCubeError)
from mlcube.parser import (CliParser, MLCubeDirectory)
from mlcube.platform import Platform
from mlcube.runner import Runner
from mlcube.shell import Shell
from mlcube.system_settings import SystemSettings
from mlcube.validate import Validate

logger = logging.getLogger(__name__)


class MultiValueOption(click.Option):
    def __init__(self, *args, **kwargs) -> None:
        super(MultiValueOption, self).__init__(*args, **kwargs)
        self._previous_parser_process: t.Optional[t.Callable] = None
        self._eat_all_parser: t.Optional[click.parser.Option] = None

    def add_to_parser(self, parser: click.parser.OptionParser, ctx: click.core.Context):

        def parser_process(value: t.Text, state: click.parser.ParsingState):
            values: t.List[t.Text] = [value]
            prefixes: t.Tuple[t.Text] = tuple(self._eat_all_parser.prefixes)
            while state.rargs:
                if state.rargs[0].startswith(prefixes):
                    break
                values.append(state.rargs.pop(0))
            self._previous_parser_process(tuple(values), state)

        super(MultiValueOption, self).add_to_parser(parser, ctx)
        for opt_name in self.opts:
            our_parser: t.Optional[click.parser.Option] = parser._long_opt.get(opt_name) or\
                                                          parser._short_opt.get(opt_name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break


def _parse_cli_args(ctx: t.Optional[click.core.Context], mlcube: t.Text, platform: t.Optional[t.Text],
                    workspace: t.Optional[t.Text],
                    resolve: bool) -> t.Tuple[t.Optional[t.Type[Runner]], DictConfig]:
    """
    Args:
        ctx: Click context. We need this to get access to extra CLI arguments.
        mlcube: Path to MLCube root directory or mlcube.yaml file.
        platform: Platform to use to run this MLCube (docker, singularity, gcp, k8s etc).
        workspace: Workspace path to use. If not specified, default workspace inside MLCube directory is used.
        resolve: if True, compute values in MLCube configuration.
    """
    mlcube_inst: MLCubeDirectory = CliParser.parse_mlcube_arg(mlcube)
    Validate.validate_type(mlcube_inst, MLCubeDirectory)
    if ctx is not None:
        mlcube_cli_args, task_cli_args = CliParser.parse_extra_arg(*ctx.args)
    else:
        mlcube_cli_args, task_cli_args = None, None
    if platform is not None:
        system_settings = SystemSettings()
        runner_config: t.Optional[DictConfig] = system_settings.get_platform(platform)
        runner_cls: t.Optional[t.Type[Runner]] = Platform.get_runner(
            system_settings.runners.get(runner_config.runner, None)
        )
    else:
        runner_cls, runner_config = None, None
    mlcube_config = MLCubeConfig.create_mlcube_config(
        os.path.join(mlcube_inst.path, mlcube_inst.file), mlcube_cli_args, task_cli_args, runner_config, workspace,
        resolve=resolve, runner_cls=runner_cls
    )
    return runner_cls, mlcube_config


log_level_option = click.option(
    '--log-level', '--log_level', required=False, type=str, default='warning',
    help="Log level to set, default is to do nothing."
)
mlcube_option = click.option(
    '--mlcube', required=False, type=str, default=os.getcwd(),
    help="Path to MLCube. This can be either a directory path that becomes MLCube's root directory, or path to MLCube"
         "definition file (.yaml). In the latter case the MLCube's root directory becomes parent directory of the yaml"
         "file. Default is current directory."
)
platform_option = click.option(
    '--platform', required=False, type=str, default='docker',
    help="Platform to run MLCube, default is 'docker' (that also supports podman)."
)
task_option = click.option(
    '--task', required=False, type=str, default=None,
    help="MLCube task name(s) to run, default is `main`. This parameter can take a list value, in which case task names"
         "are separated with ','."
)
workspace_option = click.option(
    '--workspace', required=False, type=str, default=None,
    help="Workspace location that is used to store input/output artifacts of MLCube tasks."
)


@click.group(name='mlcube', help="MLCube 📦 is a packaging tool for ML models")
@log_level_option
def cli(log_level: t.Text):
    if log_level:
        log_level = log_level.upper()
        logging.basicConfig(level=log_level)
        coloredlogs.install(level=log_level)
        logging.info("Setting Log Level from CLI argument to '%s'.", log_level)
    _ = SystemSettings().update_installed_runners()


@cli.command(name='show_config', help='Show MLCube configuration.',
             context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@mlcube_option
@platform_option
@workspace_option
@click.option('--resolve', is_flag=True, help="Resolve MLCube parameters.")
@click.pass_context
def show_config(ctx: click.core.Context, mlcube: t.Text, platform: t.Text, workspace: t.Text, resolve: bool) -> None:
    """
    Args:
        ctx: Click context. We need this to get access to extra CLI arguments.
        mlcube: Path to MLCube root directory or mlcube.yaml file.
        platform: Platform to use to run this MLCube (docker, singularity, gcp, k8s etc).
        workspace: Workspace path to use. If not specified, default workspace inside MLCube directory is used.
        resolve: if True, compute values in MLCube configuration.
    """
    _, mlcube_config = _parse_cli_args(ctx, mlcube, platform, workspace, resolve)
    print(OmegaConf.to_yaml(mlcube_config))


@cli.command(name='configure', help='Configure MLCube.',
             context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@mlcube_option
@platform_option
@click.pass_context
def configure(ctx: click.core.Context, mlcube: t.Text, platform: t.Text) -> None:
    """
    Args:
        ctx: Click context. We need this to get access to extra CLI arguments.
        mlcube: Path to MLCube root directory or mlcube.yaml file.
        platform: Platform to use to configure this MLCube for (docker, singularity, gcp, k8s etc).
    """
    runner_cls, mlcube_config = _parse_cli_args(ctx, mlcube, platform, workspace=None, resolve=True)
    docker_runner = runner_cls(mlcube_config, task=None)
    docker_runner.configure()


@cli.command(name='run', help='Run MLCube ML task.',
             context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@mlcube_option
@platform_option
@task_option
@workspace_option
@click.pass_context
def run(ctx: click.core.Context, mlcube: t.Text, platform: t.Text, task: t.Text, workspace: t.Text) -> None:
    """
    Args:
        ctx: Click context. We need this to get access to extra CLI arguments.
        mlcube: Path to MLCube root directory or mlcube.yaml file.
        platform: Platform to use to run this MLCube (docker, singularity, gcp, k8s etc).
        task: Comma separated list of tasks to run.
        workspace: Workspace path to use. If not specified, default workspace inside MLCube directory is used.
    """
    runner_cls, mlcube_config = _parse_cli_args(ctx, mlcube, platform, workspace, resolve=True)
    mlcube_tasks: t.List[str] = list((mlcube_config.get('tasks', None) or {}).keys())  # Tasks in this MLCube.
    tasks: t.List[str] = CliParser.parse_list_arg(task, default=None)                  # Requested tasks.

    if len(tasks) == 0:
        logger.warning("Missing required task name (--task=COMMA_SEPARATED_LIST_OF_TASK_NAMES).")
        if len(mlcube_tasks) != 1:
            logger.error("Task name could not be automatically resolved (supported tasks = %s).", str(mlcube_tasks))
            exit(1)
        logger.info("Task name has been automatically resolved to %s (supported tasks = %s).",
                    mlcube_tasks[0], str(mlcube_tasks))
        tasks = mlcube_tasks

    unknown_tasks: t.List[str] = [name for name in tasks if name not in mlcube_tasks]
    if len(unknown_tasks) > 0:
        logger.error("Unknown tasks have been requested: supported tasks = %s, requested tasks = %s, "
                     "unknown tasks = %s.", str(mlcube_tasks), str(tasks), str(unknown_tasks))
        exit(1)

    for task in tasks:
        logger.info("Task = %s", task)
        docker_runner = runner_cls(mlcube_config, task=task)
        docker_runner.run()


@cli.command(name='describe', help='Describe MLCube.')
@mlcube_option
def describe(mlcube: t.Text) -> None:
    _, mlcube_config = _parse_cli_args(None, mlcube, None, None, resolve=True)
    print(f"MLCube")
    print(f"  path = {mlcube_config.runtime.root}")
    print(f"  name = {mlcube_config.name}:{mlcube_config.get('version', 'latest')}")
    print()
    print(f"  workspace = {mlcube_config.runtime.workspace}")
    print(f"  system settings = {SystemSettings.system_settings_file()}")
    print()
    print(f"  Tasks:")
    for task_name, task_def in mlcube_config.tasks.items():
        description = f"name = {task_name}"
        if len(task_def.parameters.inputs) > 0:
            description = f"{description}, inputs = {list(task_def.parameters.inputs.keys())}"
        if len(task_def.parameters.outputs) > 0:
            description = f"{description}, outputs = {list(task_def.parameters.outputs.keys())}"
        print(f"    {description}")
    print()
    print(f"Run this MLCube:")
    print("  Configure MLCube:")
    print(f"    mlcube configure --mlcube={mlcube_config.runtime.root} --platform=docker")
    print("  Run MLCube tasks:")
    for task_name in mlcube_config.tasks.keys():
        print(f"    mlcube run --mlcube={mlcube_config.runtime.root} --task={task_name} --platform=docker")
    print()


@cli.command(name='config', help='Perform various operations with system settings file.',
             context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.option('--list', 'list_all', is_flag=True, help="List configuration in MLCube system settings file.")
@click.option('--get', required=False, type=str, default=None,
              help="Return value of the key (use OmegaConf notation, e.g. --get runners.docker).")
@click.option('--create_platform', '--create-platform', required=False, cls=MultiValueOption, type=tuple, default=None,
              help="Create a new platform instance for this runner")
@click.option('--remove_platform', '--remove-platform', required=False, type=str, default=None,
              help="Remove this platform from list of platforms in system settings file.")
@click.option('--rename_platform', '--rename-platform', required=False, cls=MultiValueOption, type=tuple, default=None,
              help="Rename existing platform. If default platform is to be renamed (like docker), it will be recreated "
                   "(with default values) next time mlcube runs")
@click.option('--copy_platform', '--copy-platform', required=False, cls=MultiValueOption, type=tuple, default=None,
              help="Copy existing platform.")
@click.option('--rename_runner', '--rename-runner', required=False, cls=MultiValueOption, type=tuple, default=None,
              help="Rename existing runner. If platforms exist that reference this runner, users must explicitly "
                   "provide `--update-platforms` flag to confirm they want to update platforms' description too.")
@click.option('--remove_runner', '--remove-runner', required=False, type=str, default=None,
              help="Remove existing runner from the list. If platforms exist that reference this runner, users must "
                   "explicitly provide `--remove-platforms` flag to confirm they want to remove platforms too.")
@click.pass_context
def config(ctx: click.core.Context,
           list_all: bool,                          # mlcube config --list
           get: t.Optional[t.Text],                 # mlcube config --get KEY
           create_platform: t.Optional[t.Tuple],    # mlcube config --create-platform RUNNER PLATFORM
           remove_platform: t.Optional[t.Text],     # mlcube config --remove-platform NAME
           rename_platform: t.Optional[t.Tuple],    # mlcube config --rename-platform OLD_NAME NEW_NAME
           copy_platform: t.Optional[t.Tuple],      # mlcube config --copy-platform EXISTING_PLATFORM NEW_PLATFORM
           rename_runner: t.Optional[t.Tuple],      # mlcube config --rename-runner OLD_NAME NEW_NAME
           remove_runner: t.Optional[t.Text]        # mlcube config --remove-runner NAME
           ) -> None:
    """ Work with MLCube system settings (similar to `git config`). """
    print(f"System settings file path = {SystemSettings.system_settings_file()}")
    settings = SystemSettings()

    def _check_tuple(_tuple: t.Tuple, _name: t.Text, _expected_size: int, _expected_value: t.Text) -> None:
        if len(_tuple) != _expected_size:
            raise IllegalParameterValueError(f'--{_name}', ' '.join(_tuple), f"\"{_expected_value}\"")

    try:
        if list_all:
            print(OmegaConf.to_yaml(settings.settings))
        elif get:
            print(OmegaConf.to_yaml(OmegaConf.select(settings.settings, get)))
        elif create_platform:
            _check_tuple(create_platform, 'create_platform', 2, 'RUNNER_NAME PLATFORM_NAME')
            settings.create_platform(create_platform)
        elif remove_platform:
            settings.remove_platform(remove_platform)
        elif rename_platform:
            _check_tuple(rename_platform, 'rename_platform', 2, 'OLD_NAME NEW_NAME')
            settings.copy_platform(rename_platform, delete_source=True)
        elif copy_platform:
            _check_tuple(copy_platform, 'copy_platform', 2, 'EXISTING_PLATFORM NEW_PLATFORM')
            settings.copy_platform(rename_platform, delete_source=False)
        elif rename_runner:
            _check_tuple(rename_runner, 'rename_runner', 2, 'OLD_NAME NEW_NAME')
            update_platforms: bool = '--update-platforms' in ctx.args or '--update_platforms' in ctx.args
            settings.rename_runner(rename_runner, update_platforms=update_platforms)
        elif remove_runner:
            remove_platforms: bool = '--remove-platforms' in ctx.args or '--remove_platforms' in ctx.args
            settings.remove_runner(remove_runner, remove_platforms=remove_platforms)
    except MLCubeError as e:
        logger.error("Command failed, command = '%s' error = '%s'", ' '.join(sys.argv), str(e))


@cli.command(name='create',
             help='Create a new MLCube using cookiecutter.')
def create() -> None:
    """ Create a new MLCube using cookiecutter template.
      - MLCube cookiecutter: https://github.com/mlcommons/mlcube_cookiecutter
      - Example: https://mlcommons.github.io/mlcube/tutorials/create-mlcube/
    """
    mlcube_cookiecutter_url = 'https://github.com/mlcommons/mlcube_cookiecutter'
    try:
        from cookiecutter.main import cookiecutter
        proj_dir: t.Text = cookiecutter(mlcube_cookiecutter_url)
        if proj_dir and os.path.isfile(os.path.join(proj_dir, 'mlcube.yaml')):
            Shell.run('mlcube', 'describe', '--mlcube', proj_dir)
    except ImportError:
        print("Cookiecutter library not found.")
        print("\tInstall it: pip install cookiecutter")
        print(f"\tMore details: {mlcube_cookiecutter_url}")


if __name__ == "__main__":
    cli()



import typing as t


__all__ = ['MLCubeError', 'ConfigurationError', 'IllegalParameterValueError', 'IllegalParameterTypeError']


class MLCubeError(Exception):
    """ Base class for all MLCube errors. """
    pass


class ConfigurationError(MLCubeError):
    """ Base class for all configuration errors. """
    pass


class IllegalParameterError(ConfigurationError):
    pass


class IllegalParameterValueError(IllegalParameterError):
    """ Exception to be raised when a configuration parameter is missing or has illegal value. """
    def __init__(self, name: t.Text, value: t.Any, expected_value: t.Any, namespace: t.Optional[t.Text] = None) -> None:
        self.name = namespace + '.' + name if namespace else name
        self.value = value
        self.expected_value = expected_value

        super().__init__(f"name={self.name}, actual_value={self.value}, actual_value_type={type(self.value)}, "
                         f"expected_value={self.expected_value}")


class IllegalParameterTypeError(IllegalParameterError):
    def __init__(self, name: t.Text, value: t.Any, expected_type: t.Any, namespace: t.Optional[t.Text] = None) -> None:
        self.name = namespace + '.' + name if namespace else name
        self.value = value
        self.expected_type = expected_type

        super().__init__(f"name={self.name}, actual_value={self.value}, actual_value_type={type(self.value)}, "
                         f"expected_value_type={self.expected_type}")



from click.testing import CliRunner
from mlcube.__main__ import cli

runner = CliRunner()


def test_mlcube():
    response = runner.invoke(cli)
    assert response.exit_code == 0
    assert 'Usage: mlcube [OPTIONS] COMMAND [ARGS]...' in response.output
    assert '--log-level, --log_level TEXT  Log level to set, default is to do nothing.' in response.output
    assert '--help                         Show this message and exit.' in response.output
    assert 'run          Run MLCube ML task.' in response.output
    assert 'show_config  Show MLCube configuration.' in response.output
    print("All assertions passed")



import typing as t
from omegaconf import DictConfig
from mlcube.errors import (ConfigurationError, IllegalParameterTypeError, IllegalParameterValueError)

Keys = t.Union[t.Text, t.Iterable[t.Text]]


class Validate(object):
    """ Dictionary validation utils. To be replaced by something in the future, like schema checker. """

    @staticmethod
    def format_keys(keys: t.Optional[Keys]) -> t.Iterable[t.Text]:
        """ Ensures that keys are always represented as iterables (e.g. lists). """
        if keys is None:
            return []
        if isinstance(keys, str):
            return [keys]
        return keys

    def _namespace_msg(self) -> t.Text:
        return f" Namespace = {self.namespace}." if self.namespace else ""

    def _validate_string_values(self, keys: Keys, blanks: bool = True) -> None:
        """ Check that all values of these keys are strings. """
        keys = Validate.format_keys(keys)
        for key in keys:
            value = self.config.get(key, None)
            if not isinstance(value, str):
                raise IllegalParameterTypeError(key, value, str, self.namespace)
            if not blanks and value.isspace():
                raise IllegalParameterValueError(key, value, "'non-blank string'", self.namespace)

    def __init__(self, config: DictConfig, namespace: t.Optional[t.Text]) -> None:
        self.config = config
        self.namespace = namespace or ''

    def not_none(self, keys: t.Optional[Keys] = None) -> 'Validate':
        """ Check all values are not Nones. """
        keys = Validate.format_keys(keys)
        bad_keys: t.List[t.Text] = [key for key in keys if self.config.get(key, None) is None]
        if len(bad_keys) > 0:
            raise ConfigurationError(f"Parameters must present and not be None: {bad_keys}.{self._namespace_msg()}")
        return self

    def check_unknown_keys(self, known_keys: t.Iterable[t.Text]) -> 'Validate':
        """ Check if this dict contains unknown keys. """
        unknown_keys = [key for key in self.config if key not in known_keys]
        if unknown_keys:
            raise ConfigurationError(f"Unknown keys: {unknown_keys}.{self._namespace_msg()}")
        return self

    def check_values(self, keys: Keys, type_, **kwargs) -> 'Validate':
        """ Perform value checks based upon their type. """
        if type_ is str:
            self._validate_string_values(keys, **kwargs)
        elif len(kwargs) != 0:
            raise ValueError(f"Unknown arguments ({kwargs}) for type '{type_}'")
        else:
            keys = Validate.format_keys(keys)
            for key in keys:
                value = self.config.get(key, None)
                if not isinstance(value, type_):
                    raise ConfigurationError(f"Expecting {type_} value, key={key}.{self._namespace_msg()}")
        return self

    @staticmethod
    def validate_type(obj, expected_type) -> None:
        if not isinstance(obj, expected_type):
            raise TypeError(f"Actual object type ({type(obj)}) != expected type ({expected_type}).")



import logging
import typing as t
from mlcube.errors import ConfigurationError
from omegaconf import (OmegaConf, DictConfig)


__all__ = ['RunnerConfig', 'Runner']


logger = logging.getLogger(__name__)


class RunnerConfig(object):

    DEFAULT = {}

    @staticmethod
    def merge(mlcube: DictConfig) -> None:
        ...

    @staticmethod
    def validate(mlcube: DictConfig) -> None:
        ...


RunnerConfigType = t.TypeVar('RunnerConfigType', bound='RunnerConfig')


class Runner(object):
    """ Base MLCube runner """

    CONFIG: RunnerConfigType = RunnerConfig

    def __init__(self, mlcube: t.Union[DictConfig, t.Dict], task: t.Optional[t.Text]) -> None:
        """Base runner.
        Args:
            mlcube: MLCube configuration.
            task: Task name to run.
        """
        if isinstance(mlcube, dict):
            mlcube: DictConfig = OmegaConf.create(mlcube)
        if not isinstance(mlcube, DictConfig):
            raise ConfigurationError(f"Invalid mlcube type ('{type(DictConfig)}'). Expecting 'DictConfig'.")

        self.mlcube = mlcube
        self.task = task

        logger.debug("%s configuration: %s", self.__class__.__name__, str(self.mlcube.runner))

    def configure(self) -> None:
        ...

    def run(self) -> None:
        ...



import os
import logging
import typing as t
from omegaconf import (OmegaConf, DictConfig)
from mlcube.runner import Runner

logger = logging.getLogger(__name__)

__all__ = ['IOType', 'ParameterType', 'MLCubeConfig']


class IOType(object):
    INPUT = 'input'
    OUTPUT = 'output'

    @staticmethod
    def is_valid(io: t.Text) -> bool:
        return io in (IOType.INPUT, IOType.OUTPUT)


class ParameterType(object):
    FILE = 'file'
    DIRECTORY = 'directory'
    UNKNOWN = 'unknown'

    @staticmethod
    def is_valid(io: t.Text) -> bool:
        return io in (ParameterType.FILE, ParameterType.DIRECTORY, ParameterType.UNKNOWN)


class MLCubeConfig(object):

    @staticmethod
    def ensure_values_exist(config: DictConfig, keys: t.Union[t.Text, t.List], constructor: t.Callable) -> t.List:
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if config.get(key, None) is None:
                config[key] = constructor()
        return [config[key] for key in keys]

    @staticmethod
    def get_uri(value: t.Text) -> t.Text:
        if value.startswith('storage:'):
            raise ValueError(f"Storage schema is not yet supported")
        return os.path.abspath(os.path.expanduser(value))

    @staticmethod
    def create_mlcube_config(mlcube_config_file: t.Text, mlcube_cli_args: t.Optional[DictConfig] = None,
                             task_cli_args: t.Optional[t.Dict] = None, runner_config: t.Optional[DictConfig] = None,
                             workspace: t.Optional[t.Text] = None, resolve: bool = True,
                             runner_cls: t.Optional[t.Type[Runner]] = None) -> DictConfig:
        """ Create MLCube mlcube merging different configs - base, global, local and cli.
        Args:
            mlcube_config_file: Path to mlcube.yaml file.
            mlcube_cli_args: MLCube mlcube from command line.
            task_cli_args: Task parameters from command line.
            runner_config: MLCube runner configuration, usually comes from system settings file.
            workspace: Workspace path to use in this MLCube run.
            resolve: If true, compute all values (some of them may reference other parameters or environmental
                variables).
            runner_cls: A python class for the runner type specified in `runner_config`.
        """
        if mlcube_cli_args is None:
            mlcube_cli_args = OmegaConf.create({})
        if task_cli_args is None:
            task_cli_args = {}
        if runner_config is None:
            runner_config = OmegaConf.create({})
        logger.debug("mlcube_config_file = %s", mlcube_config_file)
        logger.debug("mlcube_cli_args = %s", mlcube_cli_args)
        logger.debug("task_cli_args = %s", task_cli_args)
        logger.debug("runner_config = %s", str(runner_config))
        logger.debug("workspace = %s", workspace)

        # Load MLCube configuration and maybe override parameters from command line (like -Pdocker.build_strategy=...).
        actual_workspace = '${runtime.root}/workspace' if workspace is None else MLCubeConfig.get_uri(workspace)
        mlcube_config = OmegaConf.merge(
            OmegaConf.load(mlcube_config_file),
            mlcube_cli_args,
            OmegaConf.create({
                'runtime': {
                    'root': os.path.dirname(mlcube_config_file),
                    'workspace': actual_workspace
                },
                'runner': runner_config
            })
        )
        # Maybe this is not the best idea, but originally MLCube used $WORKSPACE token to refer to the internal
        # workspace. So, this value is here to simplify access to workspace value. BTW, in general, if files are to be
        # located inside workspace (internal or custom), users are encouraged not to use ${runtime.workspace} or
        # ${workspace} in their MLCube configuration files.
        mlcube_config['workspace'] = actual_workspace
        # Merge, for instance, docker runner config from system settings with docker config from MLCube config.
        if runner_cls:
            runner_cls.CONFIG.merge(mlcube_config)
        # Need to apply CLI arguments again just in case users provided something like -Prunner.build_strategy=...
        mlcube_config = OmegaConf.merge(mlcube_config, mlcube_cli_args)
        if runner_cls:
            runner_cls.CONFIG.validate(mlcube_config)

        for task_name in mlcube_config.tasks.keys():
            [task] = MLCubeConfig.ensure_values_exist(mlcube_config.tasks, task_name, dict)
            [parameters] = MLCubeConfig.ensure_values_exist(task, 'parameters', dict)
            [inputs, outputs] = MLCubeConfig.ensure_values_exist(parameters, ['inputs', 'outputs'], dict)

            MLCubeConfig.check_parameters(inputs, task_cli_args)
            MLCubeConfig.check_parameters(outputs, task_cli_args)

        if resolve:
            OmegaConf.resolve(mlcube_config)
        return mlcube_config

    @staticmethod
    def check_parameters(parameters: DictConfig, task_cli_args: t.Dict) -> None:
        """ Check that task parameters are defined according to MLCube schema.
        Args:
            parameters: Task parameters (`inputs` or `outputs`).
            task_cli_args: Task parameters from command line.
        This function does not set `type` of parameters (if not present) in all cases.
        """
        for name in parameters.keys():
            # The `_param_name` is anyway there, so check it's not None.
            [param_def] = MLCubeConfig.ensure_values_exist(parameters, name, dict)
            # Deal with the case when value is a string (default value).
            if isinstance(param_def, str):
                parameters[name] = {'default': param_def}
                param_def = parameters[name]
            # If `default` key is not present, use parameter name as value.
            _ = MLCubeConfig.ensure_values_exist(param_def, 'default', lambda: name)
            # One challenge is how to identify type (file, directory) of input/output parameters if users have
            # not provided these types. The below is a kind of rule-based system that tries to infer types.

            # Make sure every parameter definition contains 'type' field. Also, if it's unknown, we can assume it's a
            # directory if a value ends with forward/backward slash.
            _ = MLCubeConfig.ensure_values_exist(param_def, 'type', lambda: ParameterType.UNKNOWN)
            if param_def.type == ParameterType.UNKNOWN and param_def.default.endswith(os.sep):
                param_def.type = ParameterType.DIRECTORY
            # See if there is value on a command line
            param_def.default = task_cli_args.get(name, param_def.default)
            # Check again parameter type. Users in certain number of cases will not be providing final slash on a
            # command line for directories, so we tried to infer types above using default values. Just in case, see
            # if we can do the same with user-provided values.
            if param_def.type == ParameterType.UNKNOWN and param_def.default.endswith(os.sep):
                param_def.type = ParameterType.DIRECTORY

            # TODO: For some input parameters, that generally speaking must exist, we can figure out types later,
            #       when we actually use them (in one of the runners). One problem is when inputs are optional. In this
            #       case, we need to know their type in advance.

            # It probably does not make too much sense to see, let's say, if an input parameter exists and set its
            # type at this moment, because MLCube can run on remote hosts.



import os
import copy
import shutil
import logging
import typing as t
from pathlib import Path
from distutils import dir_util
from omegaconf import DictConfig
from mlcube.errors import ConfigurationError
from mlcube.config import (ParameterType, IOType)


__all__ = ['Shell']

logger = logging.getLogger(__name__)


class Shell(object):
    """ Helper functions to run commands. """

    @staticmethod
    def run(*cmd, die_on_error: bool = True) -> int:
        """Execute shell command.
        Args:
            cmd: Command to execute, e.g. Shell.run('ls', -lh'). This method will just join using whitespaces.
            die_on_error: If true and shell returns non-zero exit status, raise RuntimeError.
        Returns:
            Exit code.
        """
        cmd: t.Text = ' '.join(cmd)
        return_code: int = os.system(cmd)
        if return_code != 0 and die_on_error:
            logger.error("Command = '%s', return_code = %d, die_on_error = %r", cmd, return_code, die_on_error)
            raise RuntimeError("Command failed: {}".format(cmd))
        logger.info("Command = '%s', return_code = %d, die_on_error = %r", cmd, return_code, die_on_error)
        return return_code

    @staticmethod
    def docker_image_exists(docker: t.Optional[t.Text], image: t.Text) -> bool:
        """Check if docker image exists.
        Args:
            docker: Docker executable (docker/sudo docker/podman/nvidia-docker/...).
            image: Name of a docker image.
        Returns:
            True if image exists, else false.
        """
        docker = docker or 'docker'
        return Shell.run(f'{docker} inspect --type=image {image} > /dev/null 2>&1', die_on_error=False) == 0

    @staticmethod
    def ssh(connection_str: t.Text, command: t.Optional[t.Text]) -> None:
        if command:
            Shell.run('ssh', '-o', 'StrictHostKeyChecking=no', connection_str, f"'{command}'")

    @staticmethod
    def rsync_dirs(source: t.Text, dest: t.Text) -> None:
        Shell.run('rsync', '-e', "'ssh'", f"'{source}'", f"'{dest}'")

    @staticmethod
    def generate_mounts_and_args(mlcube: DictConfig, task: t.Text) -> t.Tuple[t.Dict, t.List]:
        """ Generate mount points and arguments for the give task.
        Return:
            A tuple containing two elements:
                -  A mapping from host path to path inside container.
                -  A list of task arguments.
        """
        # First task argument is always the task name.
        mounts, args = {}, [task]

        def _generate(_params: DictConfig, _io: t.Text) -> None:
            """ _params here is a dictionary containing input or output parameters.
            It maps parameter name to DictConfig(type, default)
            """
            if not IOType.is_valid(_io):
                raise ConfigurationError(f"Invalid IO = {_io}")
            for _param_name, _param_def in _params.items():
                if not ParameterType.is_valid(_param_def.type):
                    raise ConfigurationError(f"Invalid task: task={task}, param={_param_name}, "
                                             f"type={_param_def.type}. Type is invalid.")
                _host_path = Path(mlcube.runtime.workspace) / _param_def.default

                if _param_def.type == ParameterType.UNKNOWN:
                    if _io == IOType.OUTPUT:
                        raise ConfigurationError(f"Invalid task: task={task}, param={_param_name}, "
                                                 f"type={_param_def.type}. Type is unknown.")
                    else:
                        if os.path.isdir(_host_path):
                            _param_def.type = ParameterType.DIRECTORY
                        elif os.path.isfile(_host_path):
                            _param_def.type = ParameterType.FILE
                        else:
                            raise ConfigurationError(f"Invalid task: task={task}, param={_param_name}, "
                                                     f"type={_param_def.type}. Type is unknown and unable to identify "
                                                     f"it ({_host_path}).")

                if _param_def.type == ParameterType.DIRECTORY:
                    os.makedirs(_host_path, exist_ok=True)
                    mounts[_host_path] = mounts.get(
                        _host_path,
                        '/mlcube_io{}/{}'.format(len(mounts), os.path.basename(_host_path))
                    )
                    args.append('--{}={}'.format(_param_name, mounts[_host_path]))
                elif _param_def.type == ParameterType.FILE:
                    _host_path, _file_name = os.path.split(_host_path)
                    os.makedirs(_host_path, exist_ok=True)
                    new_mount = mounts.get(
                        _host_path,
                        '/mlcube_io{}/{}'.format(len(mounts), _host_path)
                    )
                    windows_match = ':\\'
                    if windows_match in new_mount:
                        index = new_mount.index(windows_match)
                        substring = new_mount[index-1:index+2]
                        new_mount = new_mount.replace(substring, '').replace('\\', '/')
                    mounts[_host_path] = new_mount
                    args.append('--{}={}'.format(_param_name, mounts[_host_path] + '/' + _file_name))

        params = mlcube.tasks[task].parameters
        _generate(params.inputs, IOType.INPUT)
        _generate(params.outputs, IOType.OUTPUT)

        return mounts, args

    @staticmethod
    def to_cli_args(args: t.Mapping[t.Text, t.Any], sep: t.Text = '=', parent_arg: t.Optional[t.Text] = None) -> t.Text:
        """ Convert dict to CLI arguments.
        Args:
            args: Dictionary with parameters.
            sep: Key-value separator. For build args and environment variables it's '=', for mount points it is ':'.
            parent_arg: If not None, a parent parameter name for each arg in args, e.g. --build-arg
        """
        parent_arg = '' if not parent_arg else parent_arg + ' '
        return ' '.join(f'{parent_arg}{k}{sep}{v}' for k, v in args.items())

    @staticmethod
    def sync_workspace(target_mlcube: DictConfig, task: t.Text) -> None:
        """
        Args:
            target_mlcube: MLCube configuration. Its name (target_) means that this configuration defines actual
                configuration where MLCube is supposed to be executed. If workspaces are different, source_mlcube will
                refer to the MLCube configuration with default (internal) workspace.
            task: Task name to be executed.
        """
        def _storage_not_supported(_uri: t.Text) -> t.Text:
            """ Helper function to guard against unsupported storage. """
            _uri = _uri.strip()
            if _uri.startswith('storage:'):
                raise NotImplementedError
            return _uri

        def _is_inside_workspace(_workspace: t.Text, _artifact: t.Text) -> bool:
            """ Check if artifact is inside this workspace. Workspace directory and artifact must exist. """
            return os.path.commonpath([_workspace]) == os.path.commonpath([_workspace, _artifact])

        def _is_ok(_parameter: t.Text, _kind: t.Text, _workspace: t.Text, _artifact: t.Text, _must_exist: bool) -> bool:
            """ Return true if this artifact needs to be synced. """
            if not _is_inside_workspace(_workspace, _artifact):
                logger.debug("[sync_workspace] task = %s, parameter = %s, artifact is not inside %s workspace "
                             "(workspace = %s, uri = %s)", task, _parameter, _kind, _workspace, _artifact)
                return False
            if _must_exist and not os.path.exists(_artifact):
                logger.debug("[sync_workspace] task = %s, parameter = %s, artifact does not exist in %s workspace "
                             "(workspace = %s, uri = %s)", task, _parameter, _kind, _workspace, _artifact)
                return False
            if not _must_exist and os.path.exists(_artifact):
                logger.debug("[sync_workspace] task = %s, parameter = %s, artifact exists in %s workspace "
                             "(workspace = %s, uri = %s)", task, _parameter, _kind, _workspace, _artifact)
                return False
            return True

        def _is_task_output(_target_artifact: t.Text, _input_parameter: t.Text) -> bool:
            """ Check of this artifact is an output of some task. """
            for _task_name, _task_def in target_mlcube.tasks.items():
                for _output_param_name, _output_param_def in _task_def.parameters.outputs.items():
                    _target_output_artifact: t.Text = Path(target_workspace) / _storage_not_supported(_output_param_def.default)

                    # Can't really use `os.path.samefile` here since files may not exist.
                    # if os.path.samefile(_target_artifact, _target_output_artifact):
                    if _target_artifact == _target_output_artifact:
                        logger.debug("[sync_workspace] task = %s, parameter = %s is an output of task = %s, "
                                     "parameter = %s", task, _input_parameter, _task_name, _output_param_name)
                        return True
            return False

        # Check if actual workspace is not internal one (which is default workspace).
        target_workspace = os.path.abspath(_storage_not_supported(target_mlcube.runtime.workspace))
        os.makedirs(target_workspace, exist_ok=True)

        source_workspace = os.path.abspath(Path(target_mlcube.runtime.root) / 'workspace')
        if not os.path.exists(source_workspace):
            logger.debug("[sync_workspace] source workspace (%s) does not exist, nothing to sync.", source_workspace)
            return
        if os.path.samefile(target_workspace, source_workspace):
            logger.debug("[sync_workspace] target workspace (%s) is the same as source workspace (%s).",
                         target_workspace, source_workspace)
            return

        if task not in target_mlcube.tasks:
            raise ValueError(f"Task does not exist: {task}")

        # Deep copy of the MLCube config with workspace set to internal workspace (we need this to resolve artifact
        # paths).
        source_mlcube: DictConfig = copy.deepcopy(target_mlcube)
        source_mlcube.runtime.workspace = source_workspace
        source_mlcube.workspace = source_workspace

        inputs: t.Mapping[t.Text, DictConfig] = target_mlcube.tasks[task].parameters.inputs
        for input_name, input_def in inputs.items():
            # TODO: add support for storage protocol. Idea is to be able to retrieve actual storage specs from
            #       system settings file. It should be possible to also specify paths within that storage (see
            #       https://en.wikipedia.org/wiki/Uniform_Resource_Identifier). For instance, the `storage:home/${name}`
            #       means that the `storage` section defines some storage labelled as `home`, and MLCube needs to use
            #       ${name} path within that storage.

            source_uri: t.Text = Path(source_workspace) / _storage_not_supported(input_def.default)

            if not _is_ok(input_name, 'source', source_workspace, source_uri, _must_exist=True):
                continue

            target_uri: t.Text = Path(target_workspace) / _storage_not_supported(input_def.default)
            if not _is_ok(input_name, 'target', target_workspace, target_uri, _must_exist=False):
                continue

            if _is_task_output(target_uri, input_name):
                continue

            if os.path.isfile(source_uri):
                os.makedirs(os.path.dirname(target_uri), exist_ok=True)
                shutil.copy(source_uri, target_uri)
            elif os.path.isdir(source_uri):
                dir_util.copy_tree(source_uri, target_uri)
            else:
                raise RuntimeError(f"Unknown artifact type (%s)", source_uri)
            logger.debug("[sync_workspace] task = %s, parameter = %s, source (%s) copied to target (%s).",
                         task, input_name, source_uri, target_uri)
